@@ -10,7 +10,6 @@ from dataclasses import dataclass
 from pathlib import Path
 from typing import Optional, Dict, Any, Tuple, List, Union, Iterable
 
-import numpy as np
 import torch
 from torch import Tensor, autocast
 from torch.cuda.amp import GradScaler
@@ -18,10 +17,13 @@ from torch.nn import Module
 from torch.nn.utils import clip_grad_norm_
 from torch.optim import Optimizer
 from torch.utils.data import DataLoader, Dataset
+from torchmetrics import Metric, MeanMetric
 from tqdm import tqdm
 
+from .functional import count_parameters
 from .logger import LoggingEngine, ClassifierType
 from .model import XZTrainerConfig, SchedulerType, LRSchedulerProtocol, CheckpointType
+from .rng import _set_rng_states, _get_rng_states
 from .sampler import ReusableSequentialSampler
 
 ModelOutputType = Union[Tensor, List]
@@ -31,14 +33,26 @@ DataType = Union[Dict[str, Any], Iterable]
 _RE_SAVE_NAME = re.compile('save-\d+\.pt')
 
 
-def _convert_model_outputs(out: ModelOutputType) -> List:
+def _detach_tensor(output: ModelOutputType, move_to_cpu: bool) -> ModelOutputType:
+    if isinstance(output, Tensor):
+        output = output.detach()
+        if move_to_cpu:
+            output = output.cpu()
+        return output
+    elif isinstance(output, List):
+        return [_detach_tensor(x, move_to_cpu) for x in output]
+    else:
+        return output
+
+
+def _convert_model_outputs_for_inference(out: ModelOutputType) -> List:
     if isinstance(out, Tensor):
         if out.ndim == 0:
-            return [out.detach().cpu()]
+            return [_detach_tensor(out, move_to_cpu=True)]
         else:
-            return [x for x in out.detach().cpu()]
+            return [_detach_tensor(x, move_to_cpu=True) for x in out]
     elif isinstance(out, List):
-        return out
+        return _detach_tensor(out, move_to_cpu=True)
     else:
         raise ValueError(f'Invalid model output type: {type(out)}')
 
@@ -73,6 +87,9 @@ class TrainContext(BaseTrainContext):
     sampler: ReusableSequentialSampler
     progress_bar: tqdm
     evaluate_data_loader: Optional[DataLoader]
+    metrics_print: Dict[str, Metric]
+    metrics_train: Dict[str, Metric]
+    metrics_evaluate: Dict[str, Metric]
 
     @property
     def total_steps_in_epoch(self) -> int:
@@ -142,17 +159,17 @@ class XZTrainable(ABC):
     ) -> Tuple[Tensor, ModelOutputsType]:
         ...
 
-    def can_stack_model_outputs(self, name: str, outputs: ModelOutputType) -> bool:
-        return False
+    @abc.abstractmethod
+    def create_metrics(self) -> Dict[str, Metric]:
+        ...
 
     @abc.abstractmethod
-    def calculate_metrics(
-            self,
-            context: BaseContext,
-            model_outputs: Dict[str, List]
-    ) -> Dict[ClassifierType, float]:
+    def update_metrics(self, model_outputs: Dict[str, List], metrics: Dict[str, Metric]):
         ...
 
+    def calculate_composition_metrics(self, metric_values: Dict[str, float]) -> Dict[str, float]:
+        return {}
+
     def on_load(self, context: TrainContext, step: int):
         pass
 
@@ -166,6 +183,15 @@ class XZTrainable(ABC):
         pass
 
 
+def _metrics_to_state_dict(metrics: Dict[str, Metric]) -> Dict[str, Dict[str, Any]]:
+    return {k: v.state_dict() for k, v in metrics.items()}
+
+
+def _load_metrics_from_state_dict(metrics: Dict[str, Metric], state_dict: Dict[str, Dict[str, Any]]):
+    for k, v in metrics.items():
+        v.load_state_dict(state_dict[k])
+
+
 class XZTrainer:
     config: XZTrainerConfig
 
@@ -190,19 +216,19 @@ class XZTrainer:
             **kwargs
         )
 
-    def _log_trainable(self, context: BaseTrainContext, model_outputs: ModelOutputsType,
-                       prev_output_lens: Optional[Dict[str, int]] = None) -> Dict[str, int]:
-        new_output_lens = {k: len(v) for k, v in model_outputs.items()}
-        if prev_output_lens is not None:
-            model_outputs = {k: v[prev_output_lens[k]:] for k, v in model_outputs.items()}
-        model_outputs = self._stack_model_outputs(model_outputs)
-        scalars = self.trainable.calculate_metrics(context, model_outputs)
-        scalars['loss'] = torch.mean(model_outputs['loss']).item()
-        for k, v in scalars.items():
+    def _calculate_reset_metrics(self, metrics: Dict[str, Metric]) -> Dict[str, float]:
+        metric_values = {}
+        for name, metric in metrics.items():
+            metric_values[name] = metric.compute()
+            metric.reset()
+        metric_values.update(self.trainable.calculate_composition_metrics(metric_values))
+        return metric_values
+
+    def _log_trainable(self, context: BaseTrainContext, metrics: Dict[str, Metric]):
+        for k, v in self._calculate_reset_metrics(metrics).items():
             context.logger.log_scalar(k, v)
         self.trainable.log(context)
         context.logger.flush()
-        return new_output_lens
 
     def _move_data_to_device(self, data: Any) -> DataType:
         if isinstance(data, Tensor):
@@ -218,10 +244,7 @@ class XZTrainer:
         else:
             return data
 
-    def _stack_model_outputs(self, outs: ModelOutputsType) -> ModelOutputsType:
-        return {k: torch.stack(v) if (k == 'loss' or self.trainable.can_stack_model_outputs(k, v)) else v for k, v in outs.items()}
-
-    def _forward_pass(self, context: BaseContext, model_outputs: Dict[str, ModelOutputType], data: DataType) -> Optional[Tuple[Tensor, ModelOutputsType]]:
+    def _forward_pass(self, context: BaseContext, data: DataType) -> Optional[Tuple[Tensor, ModelOutputsType]]:
         data = self._move_data_to_device(data)
         loss, model_output = self.trainable.step(context, data)
         if loss is not None:
@@ -229,9 +252,6 @@ class XZTrainer:
                 print('NAN loss found!')
                 if self.config.skip_nan_loss:
                     return None
-            model_outputs['loss'].append(loss.detach().cpu())
-        for k, v in model_output.items():
-            model_outputs[k].extend(_convert_model_outputs(v))
         return loss, model_output
 
     @staticmethod
@@ -246,13 +266,25 @@ class XZTrainer:
         if isinstance(context, BaseTrainContext):
             context.logger.update_top_classifier(('step', 'eval'))
 
+    def _create_metrics(self) -> Dict[str, Metric]:
+        metrics = self.trainable.create_metrics()
+        metrics['loss'] = MeanMetric()
+        for k in metrics.keys():
+            m = metrics[k]
+            m.persistent(True)
+            metrics[k] = m.to(self.device)
+        return metrics
+
+    def _update_metrics(self, loss: Tensor, model_outputs: ModelOutputsType, metrics: Dict[str, Metric]):
+        model_outputs = {k: _detach_tensor(v, move_to_cpu=False) for k, v in model_outputs.items()}
+        metrics['loss'].update(loss.detach())
+        self.trainable.update_metrics(model_outputs, metrics)
+
     def _train_epoch(self, context: TrainContext):
         self._set_training_state(context)
         context.progress_bar.update(
-            context.get_step_from_batch(context.get_actual_batch_i(0)) - 1 - context.progress_bar.n)
-
-        model_outputs = defaultdict(lambda: list())
-        prev_output_lens = defaultdict(lambda: 0)
+            context.get_step_from_batch(context.get_actual_batch_i(0)) - 1 - context.progress_bar.n
+        )
 
         for batch_i, data in enumerate(context.data_loader):
             batch_i = context.get_actual_batch_i(batch_i)
@@ -265,7 +297,9 @@ class XZTrainer:
             model_op_ctx = nullcontext() if self.config.amp_dtype is None else autocast(device_type='cuda',
                                                                                         dtype=self.config.amp_dtype)
             with model_op_ctx:
-                loss, _ = self._forward_pass(context, model_outputs, data)
+                loss, model_out = self._forward_pass(context, data)
+                self._update_metrics(loss, model_out, context.metrics_print)
+                self._update_metrics(loss, model_out, context.metrics_train)
 
             if do_update:
                 for group_i, group in enumerate(context.optimizer.param_groups):
@@ -308,17 +342,17 @@ class XZTrainer:
                 self.trainable.on_update(context, step)
 
                 if context.should_perform_step_action(self.config.print_steps, batch_i):
-                    prev_output_lens = self._log_trainable(context, model_outputs, prev_output_lens)
+                    self._log_trainable(context, context.metrics_print)
 
                 if context.evaluate_data_loader and context.should_perform_step_action(self.config.eval_steps,
                                                                                        batch_i):
                     self._set_evaluating_state(context)
                     context_eval = EvalContext.from_train_context(context)
                     with torch.no_grad():
-                        eval_model_outputs = defaultdict(lambda: list())
                         for eval_data in context_eval.data_loader:
-                            self._forward_pass(context_eval, eval_model_outputs, eval_data)
-                    self._log_trainable(context, eval_model_outputs)
+                            loss_eval, model_out_eval = self._forward_pass(context_eval, eval_data)
+                            self._update_metrics(loss_eval, model_out_eval, context.metrics_evaluate)
+                    self._log_trainable(context, context.metrics_evaluate)
                     self._set_training_state(context)
                 if context.should_perform_step_action(self.config.save_steps, batch_i):
                     self._save(context, step, batch_i)
@@ -327,7 +361,7 @@ class XZTrainer:
         if len(context.data_loader) > 0:
             context.logger.update_top_classifier(('epoch', 'train'))
             context.logger.update_time_step(context.epoch)
-            self._log_trainable(context, model_outputs)
+            self._log_trainable(context, context.metrics_train)
 
     def _cleanup_saves(self):
         if self.config.save_keep_n >= 0:
@@ -349,9 +383,10 @@ class XZTrainer:
             'epoch': context.epoch,
             'batch_i_saved_at': batch_i,
             'sampler': context.sampler.save_state(batch_i, self.config.batch_size),
-            'rng_state_torch': torch.get_rng_state(),
-            'rng_state_numpy': np.random.get_state(),
-            'rng_state_python': random.getstate()
+            'rng': _get_rng_states(),
+            'metrics_print': _metrics_to_state_dict(context.metrics_print),
+            'metrics_train': _metrics_to_state_dict(context.metrics_train),
+            'metrics_eval': _metrics_to_state_dict(context.metrics_evaluate)
         }
         torch.save(save_obj, str(save_path))
         self._cleanup_saves()
@@ -392,7 +427,6 @@ class XZTrainer:
         batches_in_epoch = self._calculate_batches_in_epoch(train_data)
         total_train_steps = self._calculate_total_steps(train_data)
 
-        print(f"Starting training experiment '{exp_name}' with total {total_train_steps} steps...")
 
         # Initialize and wrap model, optimizer and scheduler
         optim = self.config.optimizer(self.model)
@@ -408,6 +442,10 @@ class XZTrainer:
             scheduler_type = None
         model = self.model
 
+        metrics_print = self._create_metrics()
+        metrics_train = self._create_metrics()
+        metrics_eval = self._create_metrics()
+
         # Load the state
         state = self._load(resume_from)
         if state is not None:
@@ -416,12 +454,13 @@ class XZTrainer:
             if scaler is not None:
                 scaler.load_state_dict(state['scaler'])
             scheduler.load_state_dict(state['scheduler'])
-            torch.random.set_rng_state(state['rng_state_torch'].cpu())
-            np.random.set_state(state['rng_state_numpy'])
-            random.setstate(state['rng_state_python'])
+            _set_rng_states(state['rng'])
             start_from_epoch = state['epoch']
             batch_i_saved_at = state['batch_i_saved_at']
             sampler_state = state['sampler']
+            _load_metrics_from_state_dict(metrics_print, state['metrics_print'])
+            _load_metrics_from_state_dict(metrics_train, state['metrics_train'])
+            _load_metrics_from_state_dict(metrics_eval, state['metrics_eval'])
         else:
             start_from_epoch = 1
             batch_i_saved_at = -1
@@ -434,6 +473,12 @@ class XZTrainer:
         else:
             eval_dl = None
 
+        print(f"Starting training experiment '{exp_name}'...")
+        print(f'Total steps: {total_train_steps:,}')
+        print(f'Parameters [total]: {count_parameters(model):,}')
+        print(f'Parameters [train]: {count_parameters(model, lambda p: p.requires_grad):,}')
+        print(f'Parameters [fixed]: {count_parameters(model, lambda p: not p.requires_grad):,}')
+
         # Run epoch loop
         with self.config.logger.create_engine(exp_name) as logger:
             with tqdm(total=total_train_steps, desc=f'Train') as progress_bar:
@@ -471,7 +516,10 @@ class XZTrainer:
                             evaluate_data_loader=eval_dl,
                             dataset_batches=batches_in_epoch,
                             shift_batch_i=shift_batch_i,
-                            progress_bar=progress_bar
+                            progress_bar=progress_bar,
+                            metrics_print=metrics_print,
+                            metrics_train=metrics_train,
+                            metrics_evaluate=metrics_eval
                         )
                     if epoch == start_from_epoch:
                         self.trainable.on_load(context, context.get_step_from_batch(context.get_actual_batch_i(0)))
@@ -481,7 +529,12 @@ class XZTrainer:
                         scheduler.step()
         return exp_name
 
-    def load_model_checkpoint(self, checkpoint_file: str, checkpoint_type: CheckpointType):
+    def load_last_checkpoint(self):
+        save_dir = Path(self.config.save_dir) / self.config.experiment_name
+        save_files = sorted(self._get_save_files(save_dir), reverse=True)
+        self.load_model_checkpoint(checkpoint_file=save_files[0][1], checkpoint_type=CheckpointType.XZTRAINER)
+
+    def load_model_checkpoint(self, checkpoint_file: Union[str, Path], checkpoint_type: CheckpointType):
         if not Path(checkpoint_file).is_file():
             print(f"'{checkpoint_file}' file doesn't exist")
             return
@@ -508,16 +561,19 @@ class XZTrainer:
             dataset_batches=len(dataloader)
         )
         self._set_evaluating_state(context)
+        infer_metrics = self._create_metrics() if calculate_metrics else None
         with torch.no_grad():
             model_outputs = defaultdict(lambda: list())
             with tqdm(total=len(dataloader), desc=f'Inference') as progress_bar:
                 for data in dataloader:
-                    self._forward_pass(context, model_outputs, data)
+                    loss_infer, model_out_infer = self._forward_pass(context, data)
+                    for k, v in model_out_infer.items():
+                        model_outputs[k].extend(_convert_model_outputs_for_inference(v))
+                    if calculate_metrics:
+                        self._update_metrics(loss_infer, model_out_infer, infer_metrics)
                     progress_bar.update()
         self._set_training_state(context)
-        model_outputs = self._stack_model_outputs(model_outputs)
         if calculate_metrics:
-            metrics = self.trainable.calculate_metrics(context, model_outputs)
-            return model_outputs, metrics
+            return dict(model_outputs), self._calculate_reset_metrics(infer_metrics)
         else:
-            return model_outputs, {}
+            return dict(model_outputs), {}
